from argparse import _SubParsersAction, Namespace
from multiprocessing import Queue
from pydantic import BaseModel
from struct import unpack_from
from typing import Callable, NoReturn, Tuple

from ..config import Config

RunFn = Callable[[Queue, Namespace, Config], NoReturn]
SetupFn = Callable[[_SubParsersAction], Tuple[str, RunFn]]


FG_NET_FDM_VERSION = 24


class FGNetFDM(BaseModel):
    """FlightGear network communication structure from src/Network/net_fdm.hxx"""
    version: int = 0
    """increment when data values change"""
    padding: int = 0
    """padding"""
    longitude: float = 0
    """geodetic (radians)"""
    latitude: float = 0
    """geodetic (radians)"""
    altitude: float = 0
    """above sea level (meters)"""
    agl: float = 0
    """above ground level (meters)"""
    phi: float = 0
    """roll (radians)"""
    theta: float = 0
    """pitch (radians)"""
    psi: float = 0
    """yaw or true heading (radians)"""
    alpha: float = 0
    """angle of attack (radians)"""
    beta: float = 0
    """side slip angle (radians)"""
    phidot: float = 0
    """roll rate (radians/sec)"""
    thetadot: float = 0
    """pitch rate (radians/sec)"""
    psidot: float = 0
    """yaw rate (radians/sec)"""
    vcas: float = 0
    """calibrated airspeed"""
    climb_rate: float = 0
    """feet per second"""
    v_north: float = 0
    """north velocity in local/body frame, fps"""
    v_east: float = 0
    """east velocity in local/body frame, fps"""
    v_down: float = 0
    """down/vertical velocity in local/body frame, fps"""
    v_body_u: float = 0
    """ECEF velocity in body frame"""
    v_body_v: float = 0
    """ECEF velocity in body frame"""
    v_body_w: float = 0
    """ECEF velocity in body frame"""
    A_X_pilot: float = 0
    """X accel in body frame ft/sec^2"""
    A_Y_pilot: float = 0
    """Y accel in body frame ft/sec^2"""
    A_Z_pilot: float = 0
    """Z accel in body frame ft/sec^2"""
    stall_warning: float = 0
    """0.0 - 1.0 indicating the amount of stall"""
    slip_deg: float = 0
    """slip ball deflection"""
    num_engines: int = 0
    """Number of valid engines"""
    eng_state: Tuple[int, int, int, int] = (0, 0, 0, 0)
    """Engine state (off, cranking, running)"""
    rpm: Tuple[float, float, float, float] = (0, 0, 0, 0)
    """Engine RPM rev/min"""
    fuel_flow: Tuple[float, float, float, float] = (0, 0, 0, 0)
    """Fuel flow gallons/hr"""
    fuel_px: Tuple[float, float, float, float] = (0, 0, 0, 0)
    """Fuel pressure psi"""
    egt: Tuple[float, float, float, float] = (0, 0, 0, 0)
    """Exhuast gas temp deg F"""
    cht: Tuple[float, float, float, float] = (0, 0, 0, 0)
    """Cylinder head temp deg F"""
    mp_osi: Tuple[float, float, float, float] = (0, 0, 0, 0)
    """Manifold pressure"""
    tit: Tuple[float, float, float, float] = (0, 0, 0, 0)
    """Turbine Inlet Temperature"""
    oil_temp: Tuple[float, float, float, float] = (0, 0, 0, 0)
    """Oil temp deg F"""
    oil_px: Tuple[float, float, float, float] = (0, 0, 0, 0)
    """Oil pressure psi"""
    num_tanks: int = 0
    """Max number of fuel tanks"""
    fuel_quantity: Tuple[float, float, float, float] = (0, 0, 0, 0)
    num_wheels: int = 0
    wow: Tuple[int, int, int] = (0, 0, 0)
    gear_pos: Tuple[float, float, float] = (0, 0, 0)
    gear_steer: Tuple[float, float, float] = (0, 0, 0)
    gear_compression: Tuple[float, float, float] = (0, 0, 0)
    cur_time: int = 0
    """current unix time"""
    warp: int = 0
    """offset in seconds to unix time"""
    visibility: float = 0
    """visibility in meters (for env. effects)"""
    elevator: float = 0
    elevator_trim_tab: float = 0
    left_flap: float = 0
    right_flap: float = 0
    left_aileron: float = 0
    right_aileron: float = 0
    rudder: float = 0
    nose_wheel: float = 0
    speedbrake: float = 0
    spoilers: float = 0

    @classmethod
    def from_bytes(cls, data):
        version = unpack_from("!I", data)[0]
        if version != FG_NET_FDM_VERSION:
            raise ValueError(
                f"Expected protocol version {FG_NET_FDM_VERSION}, got {version}")
        this = cls()
        decoded = unpack_from(
            "!IIdddffffffffffffffffffffffIIIIIffffffffffffffffffffffffffffffffffffIffffIIIIfffffffffIifffffffffff", data)
        this.version = decoded[0]
        this.padding = decoded[1]
        this.longitude = decoded[2]
        this.latitude = decoded[3]
        this.altitude = decoded[4]
        this.agl = decoded[5]
        this.phi = decoded[6]
        this.theta = decoded[7]
        this.psi = decoded[8]
        this.alpha = decoded[9]
        this.beta = decoded[10]
        this.phidot = decoded[11]
        this.thetadot = decoded[12]
        this.psidot = decoded[13]
        this.vcas = decoded[14]
        this.climb_rate = decoded[15]
        this.v_north = decoded[16]
        this.v_east = decoded[17]
        this.v_down = decoded[18]
        this.v_body_u = decoded[19]
        this.v_body_v = decoded[20]
        this.v_body_w = decoded[21]
        this.A_X_pilot = decoded[22]
        this.A_Y_pilot = decoded[23]
        this.A_Z_pilot = decoded[24]
        this.stall_warning = decoded[25]
        this.slip_deg = decoded[26]
        this.num_engines = decoded[27]
        this.eng_state = decoded[28:32]
        this.rpm = decoded[32:36]
        this.fuel_flow = decoded[36:40]
        this.fuel_px = decoded[40:44]
        this.egt = decoded[44:48]
        this.cht = decoded[48:52]
        this.mp_osi = decoded[52:56]
        this.tit = decoded[56:60]
        this.oil_temp = decoded[60:64]
        this.oil_px = decoded[64:68]
        this.num_tanks = decoded[68]
        this.fuel_quantity = decoded[69:73]
        this.num_wheels = decoded[73]
        this.wow = decoded[74:77]
        this.gear_pos = decoded[77:80]
        this.gear_steer = decoded[80:83]
        this.gear_compression = decoded[83:86]
        this.cur_time = decoded[86]
        this.warp = decoded[87]
        this.visibility = decoded[88]
        this.elevator = decoded[89]
        this.elevator_trim_tab = decoded[90]
        this.left_flap = decoded[91]
        this.right_flap = decoded[92]
        this.left_aileron = decoded[93]
        this.right_aileron = decoded[94]
        this.rudder = decoded[95]
        this.nose_wheel = decoded[96]
        this.speedbrake = decoded[97]
        this.spoilers = decoded[98]
        return this



"""listen to UDP packets from Rotorcraft-Pilot Coupling Task simulator"""
from argparse import _SubParsersAction, ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace
from multiprocessing import Queue
import socket
from struct import unpack_from
from typing import Tuple

from ..aircraft import *
from ..config import Config
from .mytypes import RunFn


def setup(subparsers: _SubParsersAction) -> Tuple[str, RunFn]:
    NAME = 'rpctask'
    parser: ArgumentParser = subparsers.add_parser(
        NAME,
        help=__doc__,
        formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('-i', '--ip',
                        help='listen UDP adress', default='0.0.0.0')
    parser.add_argument('-p', '--port', type=int,
                        help='listen UDP port', default=5004)

    return (NAME, run)


def run(q: Queue, args: Namespace, _config: Config):

    last_trim = Controls()

    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
        # Timeout is required to handle leaving with Ctrl+C
        sock.settimeout(1.0)
        sock.bind((args.ip, args.port))
        if args.verbosity >= 0:
            print(f'Listening for UDP packets on {args.ip}:{args.port}')

        while True:
            try:
                data, _ = sock.recvfrom(1024)
                # As named in the original FlightGear protocol file
                # cyclic has coordinates 0,0 in bottom left and 1,1 in top right
                (pitch_ctrl, roll_ctrl, pwr_ctrl,
                 pitch_target, roll_target, pwr_target,
                 cyc_ftr, coll_ftr) = unpack_from('>' + 'd' * 8, data)

                state = AircraftState()
                state.ctrl = Controls()
                state.ctrl.stick_right = (roll_ctrl * 2.0) - 1.0
                state.ctrl.stick_pull = (pitch_ctrl * -2.0) + 1.0
                state.ctrl.collective_up = pwr_ctrl

                state.trgt = AircraftData()
                state.trgt.ctrl = Controls()
                state.trgt.ctrl.stick_right = (roll_target * 2.0) - 1.0
                state.trgt.ctrl.stick_pull = (pitch_target * -2.0) + 1.0
                state.trgt.ctrl.collective_up = pwr_target

                state.btn = Buttons()
                if cyc_ftr > 0.5:
                    state.btn.cyc_ftr = True
                if coll_ftr > 0.5:
                    state.btn.coll_ftr = True

                if len(data) >= 8 * 14:
                    (up_bdr_cyc, low_bdr_cyc, right_bdr_cyc, left_bdr_cyc,
                     up_bdr_coll, low_bdr_coll) = unpack_from('>' + 'd' * 6, data, 8 * 8)
                    state.brdr = Borders()
                    state.brdr.high.stick_right = (right_bdr_cyc * 2.0) - 1.0
                    state.brdr.low.stick_right = (left_bdr_cyc * 2.0) - 1.0
                    state.brdr.high.stick_pull = (low_bdr_cyc * -2.0) + 1.0
                    state.brdr.low.stick_pull = (up_bdr_cyc * -2.0) + 1.0
                    state.brdr.high.collective_up = up_bdr_coll
                    state.brdr.low.collective_up = low_bdr_coll

                if state.btn.coll_ftr:
                    last_trim.collective_up = state.ctrl.collective_up
                if state.btn.cyc_ftr:
                    last_trim.stick_right = state.ctrl.stick_right
                    last_trim.stick_pull = state.ctrl.stick_pull
                state.trim = AircraftData()
                state.trim.ctrl = last_trim

                q.put(('smol', state.smol()))
            except socket.timeout:
                continue



from enum import Enum
import os
from os import path
from typing import Dict, Optional
from pydantic import Field

from .mytypes import NestingModel


class RpctaskConfig(NestingModel):
    """Configuration for `rpctask`"""
    correct_tolerance: float = 0.03
    """Acceptable margin for green color"""
    warning_tolerance: float = 0.05
    """Acceptable margin for yellow color"""


class PfdConfig(NestingModel):
    """configuration for `pfd`"""
    ias_never_exceed: float = 167.0
    """Never exceed speed (Vne) shown on IAS tape"""
    show_ias_target: bool = True
    """Display the IAS target after receiving `trgt.instr.ias`"""
    show_vsi_target: bool = True
    """Display the VSI target after receiving `trgt.v_ned`"""
    show_adi_target: bool = True
    """Display the attitude target after receiving `trgt.att`"""
    adi_target_roll: bool = True
    """Rotate attitude target indicator to show desired roll"""
    adi_target_yaw: bool = False
    """Move attitude target indicator to show desired yaw"""
    show_flightpath: bool = True
    """Show flightpath vector (FPV) indicator on ADI"""
    show_retrograde: bool = False
    """Show reverse flight path vector on ADI"""
    move_roll_ticks: bool = False
    """Move roll angle scale with horizon, keeping sideslip triangle in place"""
    sideslip_max: float = 15.0
    """Maximal displayed sideslip angle, in degrees"""
    rpm_e_good_low: float = 0.97
    """Engine RPM safe region lower limit"""
    rpm_e_good_high: float = 1.03
    """Engine RPM safe region upper limit"""
    rpm_e_warn_low: float = 0.97
    """Engine RPM warning region lower limit"""
    rpm_e_warn_high: float = 1.03
    """Engine RPM warning region upper limit"""
    rpm_e_nominal: float = 6000.0
    """Engine RPM nominal value, in revolutions per minute"""
    rpm_r_good_low: float = 0.97
    """Rotor RPM safe region lower limit"""
    rpm_r_good_high: float = 1.03
    """Rotor RPM safe region upper limit"""
    rpm_r_warn_low: float = 0.90
    """Rotor RPM warning region lower limit"""
    rpm_r_warn_high: float = 1.10
    """Rotor RPM warning region upper limit"""
    rpm_r_nominal: float = 380.0
    """Rotor RPM nominal value, in revolutions per minute"""
    traffic_range: float = 18520.0
    """Range of displayed traffic information, in meters"""


class ApproachConfig(NestingModel):
    """Ship approach configuration"""
    nominal_alt: float = 3
    """Altitude at which the scale is 1, in meters"""
    camera_height: float = 10
    """Position of camera above aircraft origin, in meters

    Larger values of this make the scale change less drastically at low altitude"""


class InstrumentsConfig(NestingModel):
    """Configuration for instruments visualisation, units etc."""
    speed_multiplier: float = 3600.0 / 1852.0
    """Scaling factor to change state velocity in meters per second to displayed IAS and GS, default for knots"""
    altitude_multiplier: float = 1 / 0.3048
    """Scaling factor to change state altitude in meters to displayed altitude, default for feet"""
    radio_altimeter_activation: float = 2500.0 * 0.3048
    """Activation height of radio altimeter above which it is not modeled, default 2500ft"""


class CASCategory(Enum):
    """Category (severity, type) of a message shown in CAS"""
    WARNING = 1
    """Red, blinking until acknowledged"""
    CAUTION = 2
    """Amber, blinking until acknowledged"""
    ADVISORY = 3
    """Green, blinking for a fixed time"""
    STATUS = 4
    """White messages"""


class CASEvent(NestingModel):
    """Event that can be displayed in CAS"""
    category: CASCategory
    """Category to put the message into"""
    text: str
    """Text to be shown"""


class CASConfig(NestingModel):
    """Configuration for Crew Alerting System, including events"""
    events: Dict[int, CASEvent] = {}
    """Dictionary of events by their integer id"""


class Config(NestingModel):
    """Root of configuration structure

    Every config field in this and children should be provided with defaults,
    that can be overriden by config files"""

    json_schema_url: Optional[str] = Field(alias='$schema', default=None)
    """Allow the `$schema` property for specifying JSON Schema URL"""
    rpctask = RpctaskConfig()
    pfd = PfdConfig()
    approach = ApproachConfig()
    instruments = InstrumentsConfig()
    cas = CASConfig()
    start_time: Optional[float] = None
    """Epoch time in seconds of starting the program (see `time.time()`)"""


def schema_location():
    root_path = path.abspath(path.dirname(__file__))
    data_path = path.join(root_path, 'data')
    os.makedirs(data_path, exist_ok=True)
    return path.join(data_path, 'lidia-config.json')


def write_schema():
    with open(schema_location(), 'w') as out:
        out.write(Config.schema_json(by_alias=True))



#!/usr/bin/env python3

"""
Generates the Python library from newest definitions at https://github.com/marsh-sim/mavlink/tree/dialect

Doesn't accept any arguments by design.
"""

from os import path as p
import subprocess
import sys
from tempfile import TemporaryDirectory

# allow running from anywhere
root_p = p.dirname(p.dirname(__file__))

# create directory for the repository
temp_dir = TemporaryDirectory()
temp_p = p.abspath(temp_dir.name)
repo_p = p.join(temp_p, 'mavlink_repo')

# clone the mavlink repository
if not p.exists(repo_p):
    subprocess.check_call(
        ['git', 'clone', '--recursive',
            'https://github.com/marsh-sim/mavlink.git', repo_p])

# update to latest commit on the dialect branch
subprocess.check_call(['git', 'fetch'], cwd=repo_p)
subprocess.check_call(['git', 'checkout', 'origin/dialect'], cwd=repo_p)

# add repo path to import pymavlink from there
sys.path.insert(1, temp_p)
# fmt: off - don't move import to top of file
from mavlink_repo.pymavlink.generator import mavgen
from mavlink_repo.pymavlink.generator.mavparse import PROTOCOL_2_0
# fmt: on

# generate Python library with all definitions
definitions_path = p.join(repo_p, 'message_definitions', 'v1.0')
dialect_file = 'all.xml'
out_path = p.join(root_p, 'mavlink_all')

opts = mavgen.Opts(out_path, wire_protocol=PROTOCOL_2_0,
                   language='Python3', validate=True)
mavgen.mavgen(opts, [p.join(definitions_path, dialect_file)])



"""looping simulation demonstrating GUI"""
from argparse import _SubParsersAction, ArgumentDefaultsHelpFormatter, ArgumentError, ArgumentParser, Namespace
from math import pi, sin
from multiprocessing import Queue
from time import sleep, time
from typing import Tuple

from ..aircraft import *
from ..config import Config
from .mytypes import RunFn


def setup(subparsers: _SubParsersAction) -> Tuple[str, RunFn]:
    NAME = 'demo'
    parser: ArgumentParser = subparsers.add_parser(
        NAME,
        help=__doc__,
        formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('-T', '--period', type=float,
                        help='loop time in seconds', default=5.0)
    parser.add_argument('-f', '--frequency', type=float,
                        help='number of messages sent per second', default=30.0)
    parser.add_argument('--no-trim', action='store_false', dest='trim',
                        help='do not demonstrate trim function')
    parser.add_argument('--no-borders', action='store_false', dest='borders',
                        help='do not demonstrate borders function')
    parser.add_argument('--outside-range', action='store_true',
                        help='demonstrate values outside allowed range')
    parser.add_argument('--alt-zero', type=float,
                        help='altitude to hover around', default=65.0)
    parser.add_argument('--alt-change', type=float,
                        help='altitude change amplitude', default=10.0)

    return (NAME, run)


def run(q: Queue, args: Namespace, config: Config):
    if args.period <= 0:
        raise ArgumentError(args.frequency, 'period must be positive')
    if args.frequency <= 0:
        raise ArgumentError(args.frequency, 'frequency must be positive')

    outside_offset = 0.6 if args.outside_range else 0.0

    def val(phase):
        """Generate a sinusoidal value with phase offset (from 0 to 1)"""
        return 0.6 * sin((time() / args.period + phase) * 2 * pi) + outside_offset

    def cycle_index(): return int(time() // args.period) % 3
    def current_phase(): return (time() / args.period) % 1

    last_trim = Controls()
    last_v_ned = NED()
    # prevent zero division on first iteration
    last_time = time() - 1 / args.frequency

    while True:
        state = AircraftState()
        state.ned = NED()
        state.ned.down = -args.alt_zero - args.alt_change * val(0.75)

        state.att = Attitude()
        state.att.pitch = 0.5 * val(0)
        state.att.roll = 0.5 * val(0.25)
        state.att.yaw = 0.5 * val(0.5)

        state.v_body = XYZ()
        state.v_body.x = 15 + 2.5 * val(0.5)
        state.v_body.y = 1 * val(0.65)
        state.v_body.z = 2 * val(0.25)

        state.v_ned = state.xyz2ned(state.v_body)

        dt = time() - last_time
        state.a_ned = NED(down=9.81) + (state.v_ned - last_v_ned) / dt
        # a_body is simulated by instruments model if it's unset
        last_v_ned = state.v_ned
        last_time = time()

        state.ctrl = Controls()
        state.ctrl.stick_pull = val(0.1)
        state.ctrl.stick_right = val(0.35)
        state.ctrl.pedals_right = val(0.6)
        state.ctrl.collective_up = 0.3 + 0.5 * (val(0.2) + outside_offset)

        state.hrpm = HelicopterRPM()
        state.hrpm.rotor = 1.0 + 0.2 * val(0.1)
        state.hrpm.engine = 1.0 + 0.1 * val(0.0)
        # Simulate sprag clutch
        state.hrpm.rotor = max(state.hrpm.rotor, state.hrpm.engine)

        state.trgt = AircraftData()
        state.trgt.ctrl = Controls()
        state.trgt.ctrl.stick_pull = val(0.55)
        state.trgt.ctrl.stick_right = val(0.3)
        state.trgt.ctrl.pedals_right = val(0.5)
        state.trgt.ctrl.collective_up = 0.3 + 0.5 * (val(0.4) + outside_offset)
        state.trgt.att = Attitude()
        state.trgt.att.roll = 0.3 * val(0.35)
        state.trgt.att.pitch = 0.3 * val(0.1)
        state.trgt.instr = Instruments()
        state.trgt.instr.ias = 28 + 5 * val(0.6)
        state.trgt.v_body = XYZ()
        state.trgt.v_body.x = state.v_body.x = 15 + 2.0 * val(0.4)
        state.trgt.v_ned = NED()
        state.trgt.v_ned.down = 2 * val(0.6)

        state.btn = Buttons()
        if cycle_index() == 1 and args.trim:
            if current_phase() < 0.4:
                state.btn.coll_ftr = True
            if current_phase() > 0.6:
                state.btn.cyc_ftr = True

        if state.btn.coll_ftr:
            last_trim.collective_up = state.ctrl.collective_up
        if state.btn.cyc_ftr:
            last_trim.stick_right = state.ctrl.stick_right
            last_trim.stick_pull = state.ctrl.stick_pull
        state.trim = AircraftData()
        state.trim.ctrl = last_trim

        state.brdr = Borders()
        if cycle_index() == 2 and args.borders:
            if current_phase() < 0.5:
                state.brdr.low = Controls.from_list(
                    [-0.8, -0.8, 0.1, -0.8, 0.1])
                state.brdr.high = Controls.from_list(
                    [0.8, 0.8, 0.9, 0.8, 0.9])
            if current_phase() > 0.5:
                state.brdr.low = Controls.from_list(
                    [-0.5, -0.5, 0.25, -0.5, 0.25])
                state.brdr.high = Controls.from_list(
                    [0.5, 0.5, 0.75, 0.5, 0.75])

        state.model_instruments(config)
        if cycle_index() == 0:
            state.instr.gs = None
            state.instr.ralt = None
        else:
            state.instr.qnh = None

        state.set_time(config)

        state.instr.tcas = []
        state.instr.tcas.append(TrafficObject(
            vol=TrafficVolume.OTHER_TRAFFIC,
            brg=1.0, dist=12000, alt=100, vsi=-1
        ))
        state.instr.tcas.append(TrafficObject(
            vol=TrafficVolume.PROXIMATE_TRAFFIC,
            brg=0.3, dist=18000, alt=30, vsi=1
        ))
        state.instr.tcas.append(TrafficObject(
            vol=TrafficVolume.TRAFFIC_ADVISORY,
            brg=-0.3, dist=10000, alt=-10, vsi=1
        ))
        state.instr.tcas.append(TrafficObject(
            vol=TrafficVolume.RESOLUTION_ADVISORY,
            brg=-0.5, dist=5000, alt=-120, vsi=0
        ))
        # Keep traffic objects fixed in world space
        for trobj in state.instr.tcas:
            trobj.brg += state.att.yaw

        q.put(('smol', state.smol()))
        sleep(1 / args.frequency)



from argparse import Namespace
import queue
import eventlet
from msgpack import packb
from multiprocessing import Queue
from os import path
import socket
import socketio


from .config import Config


def background_loop(q: Queue, sio: socketio.Server, args: Namespace):
    udp_host, udp_port = args.passthrough_host, args.passthrough_port
    udp_sock = None
    if udp_host is not None and udp_port is not None:
        udp_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    while True:
        try:
            while not q.empty():
                event, data = q.get()
                sio.emit(event, data)
                if args.verbosity >= 2:
                    print(f"socketio.emit('{event}', {data})")
                if event == 'smol' and udp_sock is not None:
                    sent_len = udp_sock.sendto(
                        packb(data), (udp_host, udp_port))
                    if args.verbosity >= 2:
                        print(
                            f"sent {sent_len} bytes to udp:{udp_host}:{udp_port}")
        except queue.Empty:
            pass
        eventlet.sleep(0.02)


def run_server(q: Queue, args: Namespace, config: Config):
    # specifying just local path breaks when run as a module
    root_path = path.abspath(path.dirname(__file__))
    pages_path = path.join(root_path, 'pages')
    static_files = {
        # @deprecated: change to info.html
        '/': {'content_type': 'text/html', 'filename': path.join(pages_path, 'controls.html')},
        '/info': {'content_type': 'text/html', 'filename': path.join(pages_path, 'info.html')},
        '/controls': {'content_type': 'text/html', 'filename': path.join(pages_path, 'controls.html')},
        # @deprecated
        '/rpctask': {'content_type': 'text/html', 'filename': path.join(pages_path, 'controls.html')},
        '/pfd': {'content_type': 'text/html', 'filename': path.join(pages_path, 'pfd.html')},
        '/approach': {'content_type': 'text/html', 'filename': path.join(pages_path, 'approach.html')},
        '/static': path.join(root_path, 'static'),
    }

    sio = socketio.Server()

    @sio.on('config_request')
    def config_request(_sid, _environ):
        q.put(('config', config.dict()))

    app = socketio.WSGIApp(sio, static_files=static_files)
    eventlet.spawn(background_loop, q, sio, args)
    eventlet.wsgi.server(eventlet.listen((args.http_host, args.http_port)),
                         app, log_output=args.verbosity >= 3)



from enum import Enum
import json
from math import cos, sin, sqrt
from time import time
from pydantic import BaseModel
from typing import Any, Dict, List, Optional, Sequence

from .config import Config as LidiaConfig
from .mytypes import VectorModel, IntFlagModel


class Attitude(VectorModel):
    """Aircraft attitude applied in order: roll, pitch, yaw"""
    roll: float = 0
    """Rotation along longitudinal axis to the right, in radians"""
    pitch: float = 0
    """Rotation along lateral axis nose up, in radians"""
    yaw: float = 0
    """Rotation along vertical axis clockwise seen from above, in radians"""


class XYZ(VectorModel):
    """Vector in aircraft coordinate system"""
    x: float = 0
    """Longitudinal axis forward"""
    y: float = 0
    """Lateral axis to the right"""
    z: float = 0
    """Vertical axis downward"""


class NED(VectorModel):
    """Vector in local horizon coordinate system"""
    north: float = 0
    east: float = 0
    down: float = 0


class Controls(VectorModel):
    stick_right: float = 0
    """Control stick input to roll right, between -1 and 1"""
    stick_pull: float = 0
    """Control stick input to pitch up, between -1 and 1"""
    throttle: float = 0
    """Engine power input, between 0 and 1"""
    pedals_right: float = 0
    """Pedals input to yaw right, between -1 and 1"""
    collective_up: float = 0
    """Collective input to increase lift, between 0 and 1"""


class HelicopterRPM(VectorModel):
    rotor: float = 0
    """Rotor RPM, relative to nominal"""
    engine: float = 0
    """Engine RPM, relative to nominal"""


class Borders(BaseModel):
    low = Controls.from_list([-1, -1, 0, -1, 0])
    high = Controls.from_list([1, 1, 1, 1, 1])

    def smol(self) -> Dict[str, Any]:
        return {
            'low': self.low.smol(),
            'high': self.high.smol(),
        }


class Buttons(IntFlagModel):
    cyc_ftr: bool = False
    coll_ftr: bool = False


class CASMessage(BaseModel):
    msg_id: int
    blinking: bool


class CAS(BaseModel):
    """Crew Alerting System state"""
    msgs: List[CASMessage]


class TrafficVolume(Enum):
    OTHER_TRAFFIC = 0
    """No collision threat, fits in entire surveillance range; hollow white diamond"""
    PROXIMATE_TRAFFIC = 1
    """No collision threat, intruder in a specified range; filled white diamond"""
    TRAFFIC_ADVISORY = 2
    """Possible threat of collision in 40 seconds; filled amber circle"""
    RESOLUTION_ADVISORY = 3
    """Real threat of collision in 25 seconds; filled red square"""


class TrafficObject(BaseModel):
    """Object displayed on TCAS view"""
    vol: TrafficVolume = TrafficVolume.OTHER_TRAFFIC
    """Type of traffic advisory issued"""
    brg: float = 0
    """Relative bearing to right, in radians"""
    dist: float = 0
    """Distance to intruder, in meters"""
    alt: float = 0
    """Altitude difference positive for intruder higher, in meters"""
    vsi: int = 0
    """Intruder vertical speed indicator, positive for climb"""


class Instruments(BaseModel):
    ias: float = 0
    """Indicated airspeed"""
    gs: Optional[float] = None
    """Groundspeed"""
    alt: float = 0
    """Barometric altitude"""
    qnh: Optional[float] = 1013
    """Altimeter setting, None for STD"""
    ralt: Optional[float] = None
    """Radio altimeter"""
    tcas: Optional[List[TrafficObject]] = None
    """Traffic objects"""


class Transforms:
    @staticmethod
    def _matmul(matrix: Sequence[float], multiplicand: Sequence[float]) -> List[float]:
        """Multiply 3x3 matrix by 3-element vector or 3x3 matrix"""
        assert len(matrix) == 9
        assert len(multiplicand) == 3 or len(multiplicand) == 9
        if len(multiplicand) == 3:
            m = matrix
            x, y, z = multiplicand
            return [
                x * m[0] + y * m[1] + z * m[2],
                x * m[3] + y * m[4] + z * m[5],
                x * m[6] + y * m[7] + z * m[8],
            ]
        elif len(multiplicand) == 9:
            m = matrix
            n = multiplicand
            # aut
            return [
                m[0] * n[0] + m[1] * n[3] + m[2] * n[6], m[0] * n[1] + m[1] * n[4] + m[2] * n[7], m[0] * n[2] + m[1] * n[5] + m[2] * n[8],  # noqa
                m[3] * n[0] + m[4] * n[3] + m[5] * n[6], m[3] * n[1] + m[4] * n[4] + m[5] * n[7], m[3] * n[2] + m[4] * n[5] + m[5] * n[8],  # noqa
                m[6] * n[0] + m[7] * n[3] + m[8] * n[6], m[6] * n[1] + m[7] * n[4] + m[8] * n[7], m[6] * n[2] + m[7] * n[5] + m[8] * n[8],  # noqa
            ]
        else:
            raise RuntimeError('Unreachable, check assertions')

    @staticmethod
    def _transpose(matrix: Sequence[float]) -> List[float]:
        """Transpose 3x3 matrix"""
        assert len(matrix) == 9
        m = matrix
        return [
            m[0], m[3], m[6],
            m[1], m[4], m[7],
            m[2], m[5], m[8],
        ]

    @staticmethod
    def _rotmat(att: Attitude) -> List[float]:
        """Create 3x3 rotation matrix transforming body-frame vectors to outside frame"""
        s, c = sin(att.roll), cos(att.roll)
        rot_roll = [
            1, 0, 0,
            0, c, -s,
            0, s, c,
        ]
        s, c = sin(att.pitch), cos(att.pitch)
        rot_pitch = [
            c, 0, s,
            0, 1, 0,
            -s, 0, c
        ]
        s, c = sin(att.yaw), cos(att.yaw)
        rot_yaw = [
            c, -s, 0,
            s, c, 0,
            0, 0, 1
        ]

        return Transforms._matmul(rot_yaw, Transforms._matmul(rot_pitch, rot_roll))


class AircraftData(BaseModel):
    """Non-recursive state of displayed aircraft"""

    ned: Optional[NED] = None
    """Position in local horizon coordinate system, in meters"""
    att: Optional[Attitude] = None
    """Aircraft attitude, in radians"""
    v_body: Optional[XYZ] = None
    """Velocity in body frame, in meters per second"""
    v_ned: Optional[NED] = None
    """Velocity in local horizon coordinate system, in meters per second"""
    a_body: Optional[XYZ] = None
    """Acceleration measured (with gravity) in body frame, in meters per second squared"""
    a_ned: Optional[NED] = None
    """Acceleration measured in local horizon coordinate system, in meters per second squared"""
    ctrl: Optional[Controls] = None
    """Control inceptors position, normalized by max deflection"""
    hrpm: Optional[HelicopterRPM] = None
    """RPM value, normalized to nominal"""
    brdr: Optional[Borders] = None
    """Task borders for inceptors"""
    btn: Optional[Buttons] = None
    """Pressed buttons"""
    instr: Optional[Instruments] = None
    """Instrument values"""
    t_boot: Optional[int] = None
    """Time from the start of simulation, in milliseconds"""

    class Config:
        json_encoders = {
            VectorModel: VectorModel.smol,
            Borders: Borders.smol,
            IntFlagModel: IntFlagModel.smol,
        }

    def smol(self) -> Dict[str, Any]:
        """Return self as dictionary with SMOL-defined keys"""
        # HACK: JSON roundtrip is required, because there is no encoder configuration for .dict()
        return json.loads(self.json(models_as_dict=False, exclude={f for f in self.__fields__ if (getattr(self, f) is None or f in ['trgt', 'trim'])}))

    @classmethod
    def from_smol(cls, smol: dict) -> 'AircraftData':
        state = cls()
        for VectorType, name in [
            (NED, 'ned'),
            (NED, 'v_ned'),
            (NED, 'a_ned'),
            (Attitude, 'att'),
            (XYZ, 'v_body'),
            (XYZ, 'a_body'),
            (Controls, 'ctrl'),
            (HelicopterRPM, 'hrpm'),
        ]:
            if name in smol:
                setattr(state, name, VectorType.from_list(smol[name]))
        if 'brdr' in smol:
            state.brdr = Borders()
            state.brdr.low = Controls.from_list(smol['brdr']['low'])
            state.brdr.high = Controls.from_list(smol['brdr']['high'])
        if 'btn' in smol:
            state.btn = Buttons.from_list(smol['btn'])
        if 'instr' in smol:
            state.instr = Instruments()
            for name in ['ias', 'gs', 'alt', 'qnh', 'ralt']:
                if name in smol['instr']:
                    setattr(state.instr, name, smol['instr'][name])
            if 'tcas' in smol['instr']:
                state.instr.tcas = [
                    TrafficObject(**t) for t in smol['instr']['tcas']]
        if 't_boot' in smol:
            state.t_boot = smol['t_boot']

        return state

    def xyz2ned(self, vec: XYZ) -> NED:
        """Transform from body frame vector to outside frame coordinates

        Assumes no rotation if `self.att` is missing"""
        if self.att is None:
            return NED.from_list([vec.x, vec.y, vec.z])
        else:
            return NED.from_list(Transforms._matmul(Transforms._rotmat(self.att), [vec.x, vec.y, vec.z]))

    def ned2xyz(self, vec: NED) -> XYZ:
        """Transform from outside frame to body frame coordinates

        Assumes no rotation if `self.att` is missing"""
        if self.att is None:
            return XYZ.from_list([vec.north, vec.east, vec.down])
        else:
            return XYZ.from_list(Transforms._matmul(Transforms._transpose(Transforms._rotmat(self.att)), [vec.north, vec.east, vec.down]))

    def model_instruments(self, config: LidiaConfig):
        """Generate values for instruments based on known parameters and configuration"""
        self._model_ias(config)
        self._model_gs(config)
        self._model_alt(config)
        self._model_ralt(config)

        # rotate acceleration to XYZ for sideslip
        if self.a_body is None and self.a_ned is not None:
            self.a_body = self.ned2xyz(self.a_ned)

    def _get_instr(self) -> Instruments:
        if self.instr is None:
            self.instr = Instruments()
        return self.instr

    def _model_ias(self, config: LidiaConfig) -> bool:
        ias = None
        if self.v_body is not None:
            ias = self.v_body.x
        elif self.v_ned is not None and self.att is not None:
            v_body = self.ned2xyz(self.v_ned)
            ias = v_body.x

        if ias is not None:
            self._get_instr().ias = ias * config.instruments.speed_multiplier
            return True
        return False

    def _model_gs(self, config: LidiaConfig) -> bool:
        gs = None
        if self.v_ned is not None:
            gs = sqrt(self.v_ned.north ** 2 + self.v_ned.east ** 2)
        elif self.v_body is not None and self.att is not None:
            v_ned = self.xyz2ned(self.v_body)
            gs = sqrt(v_ned.north ** 2 + v_ned.east ** 2)

        if gs is not None:
            self._get_instr().gs = gs * config.instruments.speed_multiplier
            return True
        return False

    def _model_alt(self, config: LidiaConfig) -> bool:
        if self.ned is not None:
            self._get_instr().alt = -self.ned.down * config.instruments.altitude_multiplier
            return True
        return False

    def _model_ralt(self, config: LidiaConfig) -> bool:
        if self.ned is not None:
            if abs(self.ned.down) <= config.instruments.radio_altimeter_activation:
                self._get_instr().ralt = -self.ned.down * config.instruments.altitude_multiplier
                return True
        return False

    def set_time(self, config: LidiaConfig):
        if config.start_time is not None:
            self.t_boot = int((time() - config.start_time) * 1000)


class AircraftState(AircraftData):
    """Full aircraft state containing target and trim condition

    Split from `AircraftData` to prevent recursive definition"""

    trgt: Optional[AircraftData] = None
    """Target values for pilot to track"""
    trim: Optional[AircraftData] = None
    """Trim values"""

    def smol(self) -> Dict[str, Any]:
        data = super().smol()
        if self.trgt is not None:
            data['trgt'] = self.trgt.smol()
        if self.trim is not None:
            data['trim'] = self.trim.smol()
        return data

    @classmethod
    def from_smol(cls, smol: dict) -> 'AircraftState':
        state = cls(**super().from_smol(smol).dict())
        for name in ['trgt', 'trim']:
            if name in smol:
                # @deprecated
                if smol[name] is List:
                    smol[name] = {'ctrl': {name: smol[name]}}
                setattr(state, name, AircraftData.from_smol(smol[name]))

        return state


if __name__ == '__main__':
    # run this from src/ folder like this: python -m lidia.aircraft
    import os

    config = LidiaConfig()
    state = AircraftState()
    state.ned = NED.from_list([0.0, 0.0, 0.0])
    state.att = Attitude.from_list([0.0, 0.0, 0.0])
    state.v_body = XYZ.from_list([0.0, 0.0, 0.0])
    state.v_ned = NED.from_list([0.0, 0.0, 0.0])
    state.a_body = XYZ.from_list([0.0, 0.0, 0.0])
    state.a_ned = NED.from_list([0.0, 0.0, 0.0])
    state.ctrl = Controls.from_list([0.0, 0.0, 0.0, 0.0, 0.0])
    state.trgt = AircraftData()
    state.trgt.ctrl = Controls.from_list([0.0, 0.0, 0.0, 0.0, 0.0])
    state.t_boot = 0x10000000
    state.model_instruments(config)
    state.instr.tcas = []
    state.instr.tcas.append(TrafficObject(
        vol=TrafficVolume.PROXIMATE_TRAFFIC,
        brg=0.5, dist=5000.0, alt=10.0, vsi=0
    ))
    state.brdr = Borders()
    state.brdr.low = Controls.from_list([0.0, 0.0, 0.0, 0.0, 0.0])
    state.brdr.high = Controls.from_list([0.0, 0.0, 0.0, 0.0, 0.0])
    print(state.smol())
    import msgpack
    packer = msgpack.Packer(use_single_float=False)
    data_double = packer.pack(state.smol())
    packer = msgpack.Packer(use_single_float=True)
    data_float = packer.pack(state.smol())
    data = data_double[:0x21] + data_float[0x15:]
    for i, b in enumerate(data):
        print(f'0x{b:02X}, ', end='')
    print()
    os.makedirs('lidia/data', exist_ok=True)
    with open('lidia/data/aircraft.bin', 'wb') as out:
        out.write(data)
    roundtrip = msgpack.unpackb(data)
    print(roundtrip)
    print(AircraftState.from_smol(roundtrip))



"""listen to UDP packets from RPC simulator for approach task"""
from argparse import _SubParsersAction, ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace
from multiprocessing import Queue
import socket
from struct import unpack_from
from typing import Tuple

from ..aircraft import *
from ..config import Config
from .mytypes import RunFn


def setup(subparsers: _SubParsersAction) -> Tuple[str, RunFn]:
    NAME = 'approach'
    parser: ArgumentParser = subparsers.add_parser(
        NAME,
        help=__doc__,
        formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('-i', '--ip',
                        help='listen UDP adress', default='0.0.0.0')
    parser.add_argument('-p', '--port', type=int,
                        help='listen UDP port', default=5004)

    return (NAME, run)


def run(q: Queue, args: Namespace, config: Config):
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
        # Timeout is required to handle leaving with Ctrl+C
        sock.settimeout(1.0)
        sock.bind((args.ip, args.port))
        if args.verbosity >= 0:
            print(f'Listening for UDP packets on {args.ip}:{args.port}')

        while True:
            try:
                data, _ = sock.recvfrom(1024)
                # Contents of UDP packet is 4 doubles:
                # - Position of helicopter north of ship, in meters
                # - Position of helicopter east of ship, in meters
                # - Position of helicopter above  ship, in meters
                # - Helicopter heading, in radians
                #   - 0 pointing north, pi/2 (90⁰) pointing east
                (north, east, altitude, yaw) = unpack_from('>' + 'd' * 4, data)

                state = AircraftState()
                state.ned = NED()
                state.ned.north = north
                state.ned.east = east
                state.ned.down = -altitude
                state.att = Attitude()
                state.att.yaw = yaw
                state.model_instruments(config)
                state.set_time(config)

                q.put(('smol', state.smol()))
            except socket.timeout:
                continue




"""connect to MARSH Sim system, see https://marsh-sim.github.io/"""
from argparse import _SubParsersAction, ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace
import json
import msgpack
from multiprocessing import Queue
from pymavlink import mavutil
import socket
from time import time
from typing import Tuple

from ..aircraft import *
from ..config import Config
from .. import mavlink_all as mavlink
from .mytypes import RunFn


def setup(subparsers: _SubParsersAction) -> Tuple[str, RunFn]:
    NAME = 'marsh'
    parser: ArgumentParser = subparsers.add_parser(
        NAME,
        help=__doc__,
        formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('-m', '--manager',
                        help='MARSH Manager IP addr', default='127.0.0.1')

    return (NAME, run)


def run(q: Queue, args: Namespace, config: Config):

    connection_string = f'udpout:{args.manager}:24400'
    mav = mavlink.MAVLink(mavutil.mavlink_connection(connection_string))
    mav.srcSystem = 1  # default system
    mav.srcComponent = mavlink.MARSH_COMP_ID_INSTRUMENTS
    if args.verbosity >= 0:
        print(f'Connecting to MARSH Manager on {connection_string}')

    last_state = AircraftState()

    # controlling when messages should be sent
    heartbeat_next = 0.0
    heartbeat_interval = 1.0

    # monitoring connection to manager with heartbeat
    timeout_interval = 5.0
    manager_timeout = 0.0
    manager_connected = False

    # the loop goes as fast as it can, relying on the variables above for timing
    while True:
        if time() >= heartbeat_next:
            mav.heartbeat_send(
                mavlink.MAV_TYPE_GENERIC,
                mavlink.MAV_AUTOPILOT_INVALID,
                mavlink.MAV_MODE_FLAG_TEST_ENABLED,
                0,
                mavlink.MAV_STATE_ACTIVE
            )
            heartbeat_next = time() + heartbeat_interval

        state_changed = False

        # handle incoming messages
        while (message := mav.file.recv_msg()) is not None:
            message: mavlink.MAVLink_message
            if message.get_type() == 'HEARTBEAT':
                if message.get_srcComponent() == mavlink.MARSH_COMP_ID_MANAGER:
                    if not manager_connected and args.verbosity >= 0:
                        print('Connected to simulation manager')
                    manager_connected = True
                    manager_timeout = time() + timeout_interval
            elif message.get_type() == 'MANUAL_CONTROL':
                # This line helps with type hints
                mc: mavlink.MAVLink_manual_control_message = message
                state_changed = True

                last_state.ctrl = Controls()
                # Invalid axes are sent as INT16_MAX
                INVALID = 0x7FFF
                if mc.x != INVALID:
                    last_state.ctrl.stick_pull = mc.x / -1000
                if mc.y != INVALID:
                    last_state.ctrl.stick_right = mc.y / 1000
                if mc.z != INVALID:
                    last_state.ctrl.throttle = mc.z / 1000
                    last_state.ctrl.collective_up = mc.z / 1000
                if mc.z != INVALID:
                    last_state.ctrl.pedals_right = mc.r / 1000

                # TODO: assign buttons

            elif message.get_type() == 'MANUAL_SETPOINT':
                # This line helps with type hints
                ms: mavlink.MAVLink_manual_setpoint_message = message
                state_changed = True

                last_state.trgt.ctrl = Controls()
                last_state.trgt.ctrl.stick_right = ms.roll
                last_state.trgt.ctrl.stick_pull = ms.pitch
                last_state.trgt.ctrl.pedals_right = ms.yaw
                last_state.trgt.ctrl.throttle = ms.thrust
                last_state.trgt.ctrl.collective_up = ms.thrust

            elif message.get_type() == 'SIM_STATE':
                ss: mavlink.MAVLink_sim_state_message = message
                state_changed = True

                last_state.att = Attitude()
                last_state.att.roll = ss.roll
                last_state.att.pitch = ss.pitch
                last_state.att.yaw = ss.yaw

                last_state.a_body = XYZ()
                last_state.a_body.x = ss.xacc
                last_state.a_body.y = ss.yacc
                last_state.a_body.z = ss.zacc

                last_state.v_ned = NED()
                last_state.v_ned.north = ss.vn
                last_state.v_ned.east = ss.ve
                last_state.v_ned.down = ss.vd

                last_state.ned = NED()
                # TODO: Add argument for reference location, calculate north and east from that
                last_state.ned.down = -ss.alt

            elif message.get_type() == 'RAW_RPM':
                rr: mavlink.MAVLink_raw_rpm_message = message
                state_changed = True

                if last_state.hrpm is None:
                    last_state.hrpm = HelicopterRPM()

                if rr.index == 0:
                    last_state.hrpm.rotor = rr.frequency / config.pfd.rpm_r_nominal
                else:
                    last_state.hrpm.engine = rr.frequency / config.pfd.rpm_e_nominal

        if manager_connected and time() > manager_timeout:
            manager_connected = False
            if args.verbosity >= 0:
                print('Lost connection to simulation manager')

        if state_changed:
            state = last_state
            state.model_instruments(config)
            if state.trgt is not None:
                state.trgt.model_instruments(config)
            if state.trim is not None:
                state.trim.model_instruments(config)
            q.put(('smol', state.smol()))



"""mock source to show detailed help about configuration"""
from argparse import _SubParsersAction, ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace
import json
from multiprocessing import Queue
from os import path
from typing import Tuple

from ..config import Config, schema_location, write_schema
from .mytypes import RunFn


def setup(subparsers: _SubParsersAction) -> Tuple[str, RunFn]:
    NAME = 'confighelp'
    parser: ArgumentParser = subparsers.add_parser(
        NAME,
        help=__doc__,
        formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('-t', '--template', type=str,
                        help='filename to write template, ending in ".json", ".yaml" or ".toml"')
    parser.add_argument('--write-schema', action='store_true',
                        help='write JSON Schema generated from current `Config` class')

    return (NAME, run)


def run(_q: Queue, args: Namespace, config: Config):
    if not path.exists(schema_location()) or args.write_schema:
        write_schema()

    if args.template is not None:
        fname: str = args.template
        # Windows path with backward slashes doesn't work
        url = 'file:' + schema_location().replace('\\', '/')
        template = {'$schema': url}
        if fname.lower().endswith('.json'):
            with open(fname, 'w') as out:
                json.dump(template, out)
        elif fname.lower().endswith('.yaml') or fname.lower().endswith('.yml'):
            with open(fname, 'w') as out:
                out.write(
                    f'# Lidia Config schema for redhat.vscode-yaml\n# yaml-language-server: $schema={url}\n\n')
        elif fname.lower().endswith('.toml'):
            with open(fname, 'w') as out:
                # `tomli` (a.k.a. `tomllib` in >= 3.11) doesn't write toml files
                out.write(
                    f'# Lidia Config schema for tamasfe.even-better-toml\n"$schema" = "{url}"\n\n')
        else:
            raise ValueError(
                'Config files have to end with ".json", ".yaml", ".yml" or ".toml" to write correct format')

    root_path = path.abspath(path.dirname(__file__))
    help_path = path.join(root_path, 'confighelp.md')
    with open(help_path, 'r') as help_file:
        print(help_file.read())

    raise KeyboardInterrupt  # exit immediately using the usual control flow
